@@ -24,7 +24,8 @@ from Pyro4.socketserver.multiplexserver import SocketServer_Multiplex
 from Pyro4.configuration import config
 
 
-__all__ = ["URI", "Proxy", "Daemon", "current_context", "callback", "batch", "async", "expose", "behavior", "oneway", "_resolve", "_locateNS"]
+__all__ = ["URI", "Proxy", "Daemon", "current_context", "callback", "batch", "async", "expose", "behavior",
+           "oneway", "SerializedBlob", "_resolve", "_locateNS"]
 
 if sys.version_info >= (3, 0):
     basestring = str
@@ -242,6 +243,8 @@ class Proxy(object):
         self.__pyroConnLock = threading.RLock()
         util.get_serializer(config.SERIALIZER)  # assert that the configured serializer is available
         self.__async = False
+        current_context.annotations = {}
+        current_context.response_annotations = {}
 
     @property
     def _pyroHmacKey(self):
@@ -292,7 +295,10 @@ class Proxy(object):
         return super(Proxy, self).__setattr__(name, value)
 
     def __repr__(self):
-        connected = "connected" if self._pyroConnection else "not connected"
+        if self._pyroConnection:
+            connected = "connected "+self._pyroConnection.family()
+        else:
+            connected = "not connected"
         return "<%s.%s at 0x%x; %s; for %s>" % (self.__class__.__module__, self.__class__.__name__,
                                                 id(self), connected, self._pyroUri)
 
@@ -411,20 +417,25 @@ class Proxy(object):
 
     def _pyroInvoke(self, methodname, vargs, kwargs, flags=0, objectId=None):
         """perform the remote method call communication"""
+        current_context.response_annotations = {}
         with self.__pyroConnLock:
             if self._pyroConnection is None:
-                # rebind here, don't do it from inside the invoke because deadlock will occur
                 self.__pyroCreateConnection()
             serializer = util.get_serializer(self._pyroSerializer or config.SERIALIZER)
-            data, compressed = serializer.serializeCall(
-                objectId or self._pyroConnection.objectId, methodname, vargs, kwargs,
-                compress=config.COMPRESSION)
+            objectId = objectId or self._pyroConnection.objectId
+            annotations = self.__annotations()
+            if vargs and isinstance(vargs[0], SerializedBlob):
+                # special serialization of a 'blob' that stays serialized
+                data, compressed, flags = self.__serializeBlobArgs(vargs, kwargs, annotations, flags, objectId, methodname, serializer)
+            else:
+                # normal serialization of the remote call
+                data, compressed = serializer.serializeCall(objectId, methodname, vargs, kwargs, compress=config.COMPRESSION)
             if compressed:
                 flags |= message.FLAGS_COMPRESSED
             if methodname in self._pyroOneway:
                 flags |= message.FLAGS_ONEWAY
             self._pyroSeq = (self._pyroSeq + 1) & 0xffff
-            msg = message.Message(message.MSG_INVOKE, data, serializer.serializer_id, flags, self._pyroSeq, annotations=self._pyroAnnotations(), hmac_key=self._pyroHmacKey)
+            msg = message.Message(message.MSG_INVOKE, data, serializer.serializer_id, flags, self._pyroSeq, annotations=annotations, hmac_key=self._pyroHmacKey)
             if config.LOGWIRE:
                 _log_wiredata(log, "proxy wiredata sending", msg)
             try:
@@ -442,6 +453,7 @@ class Proxy(object):
                         log.error(error)
                         raise errors.SerializeError(error)
                     if msg.annotations:
+                        current_context.response_annotations = msg.annotations
                         self._pyroResponseAnnotations(msg.annotations, msg.type)
                     if self._pyroRawWireResponse:
                         msg.decompress_if_needed()
@@ -506,7 +518,7 @@ class Proxy(object):
                 if compressed:
                     flags |= message.FLAGS_COMPRESSED
                 msg = message.Message(message.MSG_CONNECT, data, serializer.serializer_id, flags, self._pyroSeq,
-                                      annotations=self._pyroAnnotations(), hmac_key=self._pyroHmacKey)
+                                      annotations=self.__annotations(False), hmac_key=self._pyroHmacKey)
                 if config.LOGWIRE:
                     _log_wiredata(log, "proxy connect sending", msg)
                 conn.send(msg.to_bytes())
@@ -546,7 +558,7 @@ class Proxy(object):
                     if replaceUri:
                         self._pyroUri = uri
                     self._pyroValidateHandshake(handshake_response)
-                    log.debug("connected to %s", self._pyroUri)
+                    log.debug("connected to %s - %s", self._pyroUri, conn.family())
                     if msg.annotations:
                         self._pyroResponseAnnotations(msg.annotations, msg.type)
                 else:
@@ -593,9 +605,7 @@ class Proxy(object):
         self._pyroMethods = set(metadata["methods"])
         self._pyroAttrs = set(metadata["attrs"])
         if log.isEnabledFor(logging.DEBUG):
-            log.debug("from meta: oneway methods=%s", sorted(self._pyroOneway))
-            log.debug("from meta: methods=%s", sorted(self._pyroMethods))
-            log.debug("from meta: attributes=%s", sorted(self._pyroAttrs))
+            log.debug("from meta: methods=%s, oneway methods=%s, attributes=%s", sorted(self._pyroMethods), sorted(self._pyroOneway), sorted(self._pyroAttrs))
         if not self._pyroMethods and not self._pyroAttrs:
             raise errors.PyroError("remote object doesn't expose any methods or attributes. Did you forget setting @expose on them?")
 
@@ -622,11 +632,9 @@ class Proxy(object):
         """returns a helper class that lets you create batched method calls on the proxy"""
         return _BatchProxyAdapter(self)
 
-    def _pyroAsync(self):
-        """returns an async version of the proxy so you can do asynchronous method calls"""
-        asyncproxy = self.__copy__()
-        asyncproxy.__async = True
-        return asyncproxy
+    def _pyroAsync(self, async=True):
+        """turns the proxy into async mode so you can do asynchronous method calls, or sets it back to normal sync mode if you set async=False."""
+        self.__async = async
 
     def _pyroInvokeBatch(self, calls, oneway=False):
         flags = message.FLAGS_BATCH
@@ -636,14 +644,9 @@ class Proxy(object):
 
     def _pyroAnnotations(self):
         """
-        Returns a dict with annotations to be sent with each message.
-        Default behavior is to include the correlation id from the current context (if it is set).
-        If you override this, don't forget to call the original method and add to the dictionary returned from it,
-        rather than simply returning a new dictionary.
-        (note that the Message may include an additional hmac annotation at the moment the message is sent)
+        Override to return a dict with custom user annotations to be sent with each request message.
+        Code using Pyro 4.56 or newer can skip this and instead set the annotations directly on the context object.
         """
-        if current_context.correlation_id:
-            return {"CORR": current_context.correlation_id.bytes}
         return {}
 
     def _pyroResponseAnnotations(self, annotations, msgtype):
@@ -651,6 +654,7 @@ class Proxy(object):
         Process any response annotations (dictionary set by the daemon).
         Usually this contains the internal Pyro annotations such as hmac and correlation id,
         and if you override the annotations method in the daemon, can contain your own annotations as well.
+        Code using Pyro 4.56 or newer can skip this and instead read the response_annotations directly from the context object.
         """
         pass
 
@@ -662,6 +666,46 @@ class Proxy(object):
         """
         return
 
+    def __annotations(self, clear=True):
+        annotations = current_context.annotations
+        if current_context.correlation_id:
+            annotations["CORR"] = current_context.correlation_id.bytes
+        else:
+            annotations.pop("CORR", None)
+        annotations.update(self._pyroAnnotations())
+        if clear:
+            current_context.annotations = {}
+        return annotations
+
+    def __serializeBlobArgs(self, vargs, kwargs, annotations, flags, objectId, methodname, serializer):
+        """
+        Special handling of a "blob" argument that has to stay serialized until explicitly deserialized in client code.
+        This makes efficient, transparent gateways or dispatchers and such possible:
+        they don't have to de/reserialize the message and are independent from the serialized class definitions.
+        Annotations are passed in because some blob metadata is added. They're not part of the blob itself.
+        """
+        if len(vargs) > 1 or kwargs:
+            raise errors.SerializeError("if SerializedBlob is used, it must be the only argument")
+        blob = vargs[0]
+        flags |= message.FLAGS_KEEPSERIALIZED
+        # Pass the objectId and methodname separately in an annotation because currently,
+        # they are embedded inside the serialized message data. And we're not deserializing that,
+        # so we have to have another means of knowing the object and method it is meant for...
+        # A better solution is perhaps to split the actual remote method arguments from the
+        # control data (object + methodname) but that requires a major protocol change.
+        # The code below is not as nice but it works without any protocol change and doesn't
+        # require a hack either - so it's actually not bad like this.
+        import marshal
+        annotations["BLBI"] = marshal.dumps((blob.info, objectId, methodname))
+        if blob._contains_blob:
+            # directly pass through the already serialized msg data from within the blob
+            protocol_msg = blob._data
+            data, compressed = protocol_msg.data, protocol_msg.flags & message.FLAGS_COMPRESSED
+        else:
+            # replaces SerializedBlob argument with the data to be serialized
+            data, compressed = serializer.serializeCall(objectId, methodname, blob._data, kwargs, compress=config.COMPRESSION)
+        return data, compressed, flags
+
 
 class _StreamResultIterator(object):
     """
@@ -672,6 +716,7 @@ class _StreamResultIterator(object):
     def __init__(self, streamId, proxy):
         self.streamId = streamId
         self.proxy = proxy
+        self.pyroseq = proxy._pyroSeq
 
     def __iter__(self):
         return self
@@ -681,16 +726,38 @@ class _StreamResultIterator(object):
         return self.__next__()
 
     def __next__(self):
+        if self.proxy is None:
+            raise StopIteration
         if self.proxy._pyroConnection is None:
             raise errors.ConnectionClosedError("the proxy for this stream result has been closed")
-        return self.proxy._pyroInvoke("get_next_stream_item", [self.streamId], {}, objectId=constants.DAEMON_NAME)
+        self.pyroseq += 1
+        try:
+            return self.proxy._pyroInvoke("get_next_stream_item", [self.streamId], {}, objectId=constants.DAEMON_NAME)
+        except (StopIteration, GeneratorExit):
+            # when the iterator is exhausted, the proxy is removed to avoid unneeded close_stream calls later
+            # (the server has closed its part of the stream by itself already)
+            self.proxy = None
+            raise
 
     def __del__(self):
         self.close()
 
     def close(self):
         if self.proxy and self.proxy._pyroConnection is not None:
-            self.proxy._pyroInvoke("close_stream", [self.streamId], {}, flags=message.FLAGS_ONEWAY, objectId=constants.DAEMON_NAME)
+            if self.pyroseq == self.proxy._pyroSeq:
+                # we're still in sync, it's okay to use the same proxy to close this stream
+                self.proxy._pyroInvoke("close_stream", [self.streamId], {}, flags=message.FLAGS_ONEWAY, objectId=constants.DAEMON_NAME)
+            else:
+                # The proxy's sequence number has diverged.
+                # One of the reasons this can happen is because this call is being done from python's GC where
+                # it decides to gc old iterator objects *during a new call on the proxy*.
+                # If we use the same proxy and do a call in between, the other call on the proxy will get an out of sync seq and crash!
+                # We create a temporary second proxy to call close_stream on. This is inefficient, but avoids the problem.
+                try:
+                    with self.proxy.__copy__() as closingProxy:
+                        closingProxy._pyroInvoke("close_stream", [self.streamId], {}, flags=message.FLAGS_ONEWAY, objectId=constants.DAEMON_NAME)
+                except errors.CommunicationError:
+                    pass
         self.proxy = None
 
 
@@ -781,6 +848,17 @@ class _AsyncRemoteMethod(object):
                 # use a copy of the proxy otherwise calls would still be done in sequence,
                 # and use contextmanager to close the proxy after we're done
                 with self.__proxy.__copy__() as proxy:
+                    delay = 0.1 + random.random()/5
+                    while not proxy._pyroConnection:
+                        try:
+                            proxy._pyroBind()
+                        except errors.CommunicationError as x:
+                            if "no free workers" not in str(x):
+                                raise
+                            time.sleep(delay)   # wait a bit until a worker might be available again
+                            delay += 0.4 + random.random()/2
+                            if 0 < config.COMMTIMEOUT/2 < delay:
+                                raise
                     value = proxy._pyroInvoke(self.__name, args, kwargs)
                 asyncresult.value = value
                 return
@@ -801,9 +879,9 @@ def batch(proxy):
     return proxy._pyroBatch()
 
 
-def async(proxy):
-    """convenience method to get an async proxy"""
-    return proxy._pyroAsync()
+def async(proxy, async=True):
+    """convenience method to set proxy to async or sync mode."""
+    proxy._pyroAsync(async)
 
 
 def pyroObjectToAutoProxy(obj):
@@ -943,6 +1021,7 @@ class DaemonObject(object):
         try:
             return next(stream)
         except Exception:
+            # in case of error (or StopIteration!) the stream is removed
             del self.daemon.streaming_responses[streamId]
             raise
 
@@ -974,10 +1053,14 @@ class Daemon(object):
             self.transportServer = SocketServer_Multiplex()
         else:
             raise errors.PyroError("invalid server type '%s'" % config.SERVERTYPE)
+        self.__mustshutdown = threading.Event()
+        self.__mustshutdown.set()
+        self.__loopstopped = threading.Event()
+        self.__loopstopped.set()
         self.transportServer.init(self, host, port, unixsocket)
         #: The location (str of the form ``host:portnumber``) on which the Daemon is listening
         self.locationStr = self.transportServer.locationStr
-        log.debug("created daemon on %s (pid %d)", self.locationStr, os.getpid())
+        log.debug("daemon created on %s - %s (pid %d)", self.locationStr, socketutil.family_str(self.transportServer.sock), os.getpid())
         natport_for_loc = natport
         if natport == 0:
             # expose internal port number as NAT port as well. (don't use port because it could be 0 and will be chosen by the OS)
@@ -990,9 +1073,6 @@ class Daemon(object):
         pyroObject._pyroId = constants.DAEMON_NAME
         #: Dictionary from Pyro object id to the actual Pyro object registered by this id
         self.objectsById = {pyroObject._pyroId: pyroObject}
-        self.__mustshutdown = threading.Event()
-        self.__loopstopped = threading.Event()
-        self.__loopstopped.set()
         # assert that the configured serializers are available, and remember their ids:
         self.__serializer_ids = {util.get_serializer(ser_name).serializer_id for ser_name in config.SERIALIZERS_ACCEPTED}
         log.debug("accepted serializers: %s" % config.SERIALIZERS_ACCEPTED)
@@ -1001,6 +1081,7 @@ class Daemon(object):
         self._pyroInstances = {}   # pyro objects for instance_mode=single (singletons, just one per daemon)
         self.streaming_responses = {}   # stream_id -> (client, creation_timestamp, linger_timestamp, stream)
         self.housekeeper_lock = threading.Lock()
+        self.__mustshutdown.clear()
 
     @property
     def _pyroHmacKey(self):
@@ -1149,7 +1230,7 @@ class Daemon(object):
             flags = message.FLAGS_COMPRESSED if compressed else 0
         # We need a minimal amount of response data or the socket will remain blocked
         # on some systems... (messages smaller than 40 bytes)
-        msg = message.Message(msgtype, data, serializer_id, flags, msg_seq, annotations=self.annotations(), hmac_key=self._pyroHmacKey)
+        msg = message.Message(msgtype, data, serializer_id, flags, msg_seq, annotations=self.__annotations(), hmac_key=self._pyroHmacKey)
         if config.LOGWIRE:
             _log_wiredata(log, "daemon handshake response", msg)
         conn.send(msg.to_bytes())
@@ -1191,15 +1272,12 @@ class Daemon(object):
             request_flags = msg.flags
             request_seq = msg.seq
             request_serializer_id = msg.serializer_id
-            if "CORR" in msg.annotations:
-                current_context.correlation_id = uuid.UUID(bytes=msg.annotations["CORR"])
-            else:
-                current_context.correlation_id = uuid.uuid4()
+            current_context.correlation_id = uuid.UUID(bytes=msg.annotations["CORR"]) if "CORR" in msg.annotations else uuid.uuid4()
             if config.LOGWIRE:
                 _log_wiredata(log, "daemon wiredata received", msg)
             if msg.type == message.MSG_PING:
                 # return same seq, but ignore any data (it's a ping, not an echo). Nothing is deserialized.
-                msg = message.Message(message.MSG_PING, b"pong", msg.serializer_id, 0, msg.seq, annotations=self.annotations(), hmac_key=self._pyroHmacKey)
+                msg = message.Message(message.MSG_PING, b"pong", msg.serializer_id, 0, msg.seq, annotations=self.__annotations(), hmac_key=self._pyroHmacKey)
                 if config.LOGWIRE:
                     _log_wiredata(log, "daemon wiredata sending", msg)
                 conn.send(msg.to_bytes())
@@ -1207,7 +1285,12 @@ class Daemon(object):
             if msg.serializer_id not in self.__serializer_ids:
                 raise errors.SerializeError("message used serializer that is not accepted: %d" % msg.serializer_id)
             serializer = util.get_serializer_by_id(msg.serializer_id)
-            objId, method, vargs, kwargs = serializer.deserializeCall(msg.data, compressed=msg.flags & message.FLAGS_COMPRESSED)
+            if request_flags & message.FLAGS_KEEPSERIALIZED:
+                # pass on the wire protocol message blob unchanged
+                objId, method, vargs, kwargs = self.__deserializeBlobArgs(msg)
+            else:
+                # normal deserialization of remote call arguments
+                objId, method, vargs, kwargs = serializer.deserializeCall(msg.data, compressed=msg.flags & message.FLAGS_COMPRESSED)
             current_context.client = conn
             current_context.client_sock_addr = conn.sock.getpeername()   # store this because on oneway calls the socket will be disconnected
             current_context.seq = msg.seq
@@ -1275,7 +1358,9 @@ class Daemon(object):
                     response_flags |= message.FLAGS_COMPRESSED
                 if wasBatched:
                     response_flags |= message.FLAGS_BATCH
-                msg = message.Message(message.MSG_RESULT, data, serializer.serializer_id, response_flags, request_seq, annotations=self.annotations(), hmac_key=self._pyroHmacKey)
+                msg = message.Message(message.MSG_RESULT, data, serializer.serializer_id, response_flags, request_seq,
+                                      annotations=self.__annotations(), hmac_key=self._pyroHmacKey)
+                current_context.response_annotations = {}
                 if config.LOGWIRE:
                     _log_wiredata(log, "daemon wiredata sending", msg)
                 conn.send(msg.to_bytes())
@@ -1286,7 +1371,8 @@ class Daemon(object):
                 request_seq = msg.seq
                 request_serializer_id = msg.serializer_id
             if xt is not errors.ConnectionClosedError:
-                log.debug("Exception occurred while handling request: %r", xv)
+                if xt not in (StopIteration, GeneratorExit):
+                    log.debug("Exception occurred while handling request: %r", xv)
                 if not request_flags & message.FLAGS_ONEWAY:
                     if isinstance(xv, errors.SerializeError) or not isinstance(xv, errors.CommunicationError):
                         # only return the error to the client if it wasn't a oneway call, and not a communication error
@@ -1386,7 +1472,7 @@ class Daemon(object):
         else:
             raise errors.DaemonError("invalid instancemode in registered class")
 
-    def _sendExceptionResponse(self, connection, seq, serializer_id, exc_value, tbinfo, flags=0, annotations={}):
+    def _sendExceptionResponse(self, connection, seq, serializer_id, exc_value, tbinfo, flags=0, annotations=None):
         """send an exception back including the local traceback info"""
         exc_value._pyroTraceback = tbinfo
         if sys.platform == "cli":
@@ -1406,9 +1492,9 @@ class Daemon(object):
         flags |= message.FLAGS_EXCEPTION
         if compressed:
             flags |= message.FLAGS_COMPRESSED
-        ann = self.annotations()
-        ann.update(annotations or {})
-        msg = message.Message(message.MSG_RESULT, data, serializer.serializer_id, flags, seq, annotations=ann, hmac_key=self._pyroHmacKey)
+        annotations = dict(annotations or {})
+        annotations.update(self.annotations())
+        msg = message.Message(message.MSG_RESULT, data, serializer.serializer_id, flags, seq, annotations=annotations, hmac_key=self._pyroHmacKey)
         if config.LOGWIRE:
             _log_wiredata(log, "daemon wiredata sending (error response)", msg)
         connection.send(msg.to_bytes())
@@ -1532,14 +1618,7 @@ class Daemon(object):
             self.transportServer = None
 
     def annotations(self):
-        """
-        Returns a dict with annotations to be sent with each message.
-        Default behavior is to include the correlation id from the current context (if it is set).
-        If you override this, don't forget to call the original method and add to the dictionary returned from it,
-        rather than simply returning a new dictionary.
-        """
-        if current_context.correlation_id:
-            return {"CORR": current_context.correlation_id.bytes}
+        """Override to return a dict with custom user annotations to be sent with each response message."""
         return {}
 
     def combine(self, daemon):
@@ -1551,10 +1630,20 @@ class Daemon(object):
         log.debug("combining event loop with other daemon")
         self.transportServer.combine_loop(daemon.transportServer)
 
+    def __annotations(self):
+        annotations = current_context.response_annotations
+        if current_context.correlation_id:
+            annotations["CORR"] = current_context.correlation_id.bytes
+        else:
+            annotations.pop("CORR", None)
+        annotations.update(self.annotations())
+        return annotations
+
     def __repr__(self):
         if hasattr(self, "locationStr"):
-            return "<%s.%s at 0x%x; %s; %d objects>" % (self.__class__.__module__, self.__class__.__name__,
-                                                        id(self), self.locationStr, len(self.objectsById))
+            family = socketutil.family_str(self.sock)
+            return "<%s.%s at 0x%x; %s - %s; %d objects>" % (self.__class__.__module__, self.__class__.__name__,
+                                                             id(self), self.locationStr, family, len(self.objectsById))
         else:
             # daemon objects may come back from serialized form without being properly initialized (by design)
             return "<%s.%s at 0x%x; unusable>" % (self.__class__.__module__, self.__class__.__name__, id(self))
@@ -1594,6 +1683,16 @@ class Daemon(object):
             return True, None
         return False, data
 
+    def __deserializeBlobArgs(self, protocolmsg):
+        import marshal
+        blobinfo = protocolmsg.annotations["BLBI"]
+        if sys.platform == "cli" and type(blobinfo) is not str:
+            # Ironpython's marshal expects str...
+            blobinfo = str(blobinfo)
+        blobinfo, objId, method = marshal.loads(blobinfo)
+        blob = SerializedBlob(blobinfo, protocolmsg, is_blob=True)
+        return objId, method, (blob,), {}  # object, method, vargs, kwargs
+
 
 # serpent serializer initialization
 
@@ -1645,6 +1744,7 @@ class _CallContext(threading.local):
         self.msg_flags = 0
         self.serializer_id = 0
         self.annotations = {}
+        self.response_annotations = {}
         self.correlation_id = None
 
     def to_global(self):
@@ -1657,6 +1757,7 @@ class _CallContext(threading.local):
             "msg_flags": self.msg_flags,
             "serializer_id": self.serializer_id,
             "annotations": self.annotations,
+            "response_annotations": self.response_annotations,
             "correlation_id": self.correlation_id,
             "client_sock_addr": self.client_sock_addr
         }
@@ -1667,6 +1768,7 @@ class _CallContext(threading.local):
         self.msg_flags = values["msg_flags"]
         self.serializer_id = values["serializer_id"]
         self.annotations = values["annotations"]
+        self.response_annotations = values["response_annotations"]
         self.correlation_id = values["correlation_id"]
         self.client_sock_addr = values["client_sock_addr"]
 
@@ -1741,6 +1843,8 @@ def _locateNS(host=None, port=None, broadcast=True, hmac_key=None):
                     return proxy
                 except errors.PyroError:
                     pass
+        if config.PREFER_IP_VERSION == 6:
+            broadcast = False   # ipv6 doesn't have broadcast. We should probably use multicast....
         if broadcast:
             # broadcast lookup
             if not port:
@@ -1803,6 +1907,34 @@ def _locateNS(host=None, port=None, broadcast=True, hmac_key=None):
         raise e
 
 
+class SerializedBlob(object):
+    """
+    Used to wrap some data to make Pyro pass this object transparently (it keeps the serialized payload as-is)
+    Only when you need to access the actual client data you can deserialize on demand.
+    This makes efficient, transparent gateways or dispatchers and such possible:
+    they don't have to de/reserialize the message and are independent from the serialized class definitions.
+    You have to pass this as the only parameter to a remote method call for Pyro to understand it.
+    Init arguments:
+    ``info`` = some (small) descriptive data about the blob. Can be a simple id or name or guid. Must be marshallable.
+    ``data`` = the actual client data payload that you want to transfer in the blob. Can be anything that you would
+    otherwise have used as regular remote call arguments.
+    """
+    def __init__(self, info, data, is_blob=False):
+        self.info = info
+        self._data = data
+        self._contains_blob = is_blob
+
+    def deserialized(self):
+        """Retrieves the client data stored in this blob. Deserializes the data automatically if required."""
+        if self._contains_blob:
+            protocol_msg = self._data
+            serializer = util.get_serializer_by_id(protocol_msg.serializer_id)
+            _, _, data, _ = serializer.deserializeData(protocol_msg.data, protocol_msg.flags & message.FLAGS_COMPRESSED)
+            return data
+        else:
+            return self._data
+
+
 # call context thread local
 
 current_context = _CallContext()
